import asyncio

import pytest
from redis.asyncio import Redis
from testcontainers.redis import RedisContainer

from redguard import RedGuard


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.stop()


@pytest.fixture(scope="session")
def redis():
    with RedisContainer() as container:
        url = f"redis://{container.get_container_host_ip()}:{container.get_exposed_port(container.port)}/0"
        client = Redis.from_url(url)
        yield client


@pytest.fixture(scope="session")
def guard(redis):
    return RedGuard(redis, namespace="test")



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING
from uuid import uuid4

from .abc import SynchronizationPrimitive

if TYPE_CHECKING:
    from redis.asyncio import Redis
    from redis.commands.core import AsyncScript


class Semaphore(SynchronizationPrimitive):
    """
    Distributed semaphore.
    Allows up to 'capacity' operations at the same time.
    timeout is required to avoid deadlocks
    """

    LUA_SCRIPT = """
    local key = KEYS[1]
    local capacity = tonumber(ARGV[1])
    local token = ARGV[2]
    local ttl = tonumber(ARGV[3])
    local current_time = redis.call('TIME')
    local current_time_micro = tonumber(current_time[1]) * 1000000 + tonumber(current_time[2])
    local token_expiration = current_time_micro + (ttl  * 1000000)

    redis.call('ZREMRANGEBYSCORE', key, 0, current_time_micro)
    if redis.call('ZSCORE', key, token) then
        return true -- token already acquired so return true
    end

    local current_count = redis.call('ZCARD', key)

    if current_count < capacity then
        redis.call('ZADD', key, token_expiration, token)

        if current_count == 0 then
            redis.call('PEXPIRE', key, ttl * 1000)
        end

        return true
    end
    return false
    """

    script: AsyncScript

    def __init__(
        self,
        client: Redis,
        name: str,
        *,
        capacity: int,
        ttl: int,
    ):
        super().__init__(client, name)
        if capacity <= 1:
            raise ValueError("capacity must be greater than 1")
        if ttl <= 0:
            raise ValueError("ttl must be greater than 1")
        self.capacity = capacity
        self.ttl = ttl
        self.tokens: deque[bytes] = deque()

        if not hasattr(self, "script"):
            type(self).script = self.client.register_script(self.LUA_SCRIPT)

    async def maybe_acquire(self) -> bool:
        token = uuid4().bytes
        result = await self.script([self.name], [self.capacity, token, self.ttl])
        acquired = bool(result)
        if acquired:
            self.tokens.append(token)
        return acquired

    async def release(self) -> None:
        if len(self.tokens) > 0:
            token = self.tokens.popleft()
            await self.client.zrem(self.name, token)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from .abc import SynchronizationPrimitive

T = TypeVar("T")


class SharedResourcePool(Generic[T]):
    def __init__(
        self,
        primitive: SynchronizationPrimitive,
        *,
        factory: Callable[[], T],
        timeout: float | None = None,
    ) -> None:
        self.primitive = primitive
        self.factory = factory
        self.timeout = timeout

    async def acquire(self) -> T:
        await self.primitive.acquire(blocking=True, timeout=self.timeout)
        return self.factory()

    async def release(self) -> None:
        await self.primitive.release()

    async def __aenter__(self) -> T:
        return await self.acquire()

    async def __aexit__(self, *exc) -> None:
        await self.release()



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import uuid4

from .abc import SynchronizationPrimitive

if TYPE_CHECKING:
    from redis.asyncio import Redis
    from redis.commands.core import AsyncScript


class Lock(SynchronizationPrimitive):
    """
    Distributed mutex.
    Grants exclusive access to resource.
    """

    LUA_RELEASE_SCRIPT = """
    local key = KEYS[1]
    local client_token = ARGV[1]
    local actual_token = redis.call('GET', key)
    if actual_token and actual_token == client_token then
        redis.call('DEL', key)
    end
    """

    script: AsyncScript

    def __init__(
        self,
        client: Redis,
        name: str,
        *,
        ttl: int,
    ) -> None:
        super().__init__(client, name)
        self.ttl = ttl * 1000
        self.token: bytes | None = None

        if not hasattr(self, "script"):
            type(self).script = self.client.register_script(self.LUA_RELEASE_SCRIPT)

    async def maybe_acquire(self) -> bool:
        token = uuid4().bytes
        result = await self.client.set(self.name, token, nx=True, px=self.ttl)
        acquired = bool(result)
        if acquired:
            self.token = token
        return acquired

    async def release(self) -> None:
        if self.token:
            await self.script([self.name], [self.token])



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import uuid4

from .abc import SynchronizationPrimitive

if TYPE_CHECKING:
    from redis.asyncio import Redis
    from redis.commands.core import AsyncScript


class RateLimiter(SynchronizationPrimitive):
    """
    Rate limiter implementing sliding window algorithm.
    Allows up to limit operations over window (in seconds) period.
    """

    LUA_SCRIPT = """
    local key = KEYS[1]
    local token = ARGV[1]
    local limit = tonumber(ARGV[2])
    local window_size = tonumber(ARGV[3])
    local current_time = redis.call('TIME')
    local current_time_micro = tonumber(current_time[1]) * 1000000 + tonumber(current_time[2])
    local trim_time_micro = current_time_micro - (window_size  * 1000000)

    redis.call('ZREMRANGEBYSCORE', key, 0, trim_time_micro)
    local current_count = redis.call('ZCARD', key)

    if current_count < limit then
        redis.call('ZADD', key, current_time_micro, token)

        if current_count == 0 then
            redis.call('PEXPIRE', key, window_size * 1000)
        end

        return true
    end
    return false
    """

    script: AsyncScript

    def __init__(
        self, client: Redis, name: str, *, limit: int, window: int = 1
    ) -> None:
        super().__init__(client, name)
        if window <= 0:
            raise ValueError("Window must be greater than 0")
        if limit < 1:
            raise ValueError("Limit must be greater than 0")
        self.window = window
        self.limit = limit

        if not hasattr(self, "script"):
            type(self).script = self.client.register_script(self.LUA_SCRIPT)

    async def maybe_acquire(self) -> bool:
        token = uuid4().bytes
        result = await self.script([self.name], [token, self.limit, self.window])
        return bool(result)

    async def release(self) -> None:
        pass



from __future__ import annotations

from typing import Any, Callable, TypeVar

from redis.asyncio import Redis
from typing_extensions import Self

from .abc import SynchronizationPrimitive
from .limiter import RateLimiter
from .lock import Lock
from .pool import SharedResourcePool, T
from .semaphore import Semaphore

P = TypeVar("P", bound=SynchronizationPrimitive)


class RedGuard:
    """
    Helper class implementing factory pattern for synchronization primitives
    """

    def __init__(self, client: Redis, namespace: str | None = None) -> None:
        self.client = client
        self.namespace = namespace

    @classmethod
    def from_url(cls, url: str, namespace: str | None = None, **client_options) -> Self:
        return cls(Redis.from_url(url, **client_options), namespace=namespace)

    def format(self, name: str) -> str:
        if self.namespace is None:
            return name
        return f"{self.namespace}:{name}"

    def new(self, cls: type[P], name: str, **kwargs: Any) -> P:
        return cls(self.client, self.format(name), **kwargs)

    def lock(self, name: str, *, ttl: int) -> Lock:
        return self.new(Lock, name, ttl=ttl)

    def semaphore(
        self,
        name: str,
        *,
        capacity: int,
        ttl: int,
    ) -> Semaphore:
        return self.new(Semaphore, name, capacity=capacity, ttl=ttl)

    def limiter(self, name: str, *, limit: int, window: int = 1) -> RateLimiter:
        return self.new(RateLimiter, name, limit=limit, window=window)

    def pool(
        self,
        cls: type[P],
        name: str,
        factory: Callable[[], T],
        timeout: float | None = None,
        **kwargs: Any,
    ) -> SharedResourcePool[T]:
        return SharedResourcePool(
            self.new(cls, name, **kwargs), factory=factory, timeout=timeout
        )



import asyncio

from redguard import Lock, RateLimiter, RedGuard, Semaphore


async def test_lock(guard: RedGuard):
    lock = guard.new(Lock, "test-lock", ttl=10)
    acquired = await lock.acquire()
    assert acquired is True
    acquired = await lock.acquire()
    assert acquired is False
    await lock.release()
    acquired = await lock.acquire()
    assert acquired is True
    await lock.release()


async def test_semaphore(guard: RedGuard):
    capacity = 3
    semaphore = guard.new(Semaphore, "test-sem", capacity=capacity, ttl=10)

    for i in range(capacity + 1):
        acquired = await semaphore.acquire()
        assert acquired is (i < capacity)

    for _ in range(capacity + 1):
        await semaphore.release()


async def test_rate_limiter(guard: RedGuard):
    await asyncio.sleep(1)
    limiter = guard.new(RateLimiter, "test-limiter", limit=3, window=1)
    for _ in range(3):
        assert await limiter.acquire() is True
    assert await limiter.acquire() is False

    await limiter.release()


async def test_object_pool(guard: RedGuard):
    def dict_factory() -> dict[str, str]:
        return {}

    pool = guard.pool(Lock, "text-pool", factory=dict_factory, ttl=1)

    async with pool as resource:
        assert isinstance(resource, dict)



from .__about__ import __version__
from .guard import RedGuard
from .limiter import RateLimiter
from .lock import Lock
from .pool import SharedResourcePool
from .semaphore import Semaphore

__all__ = [
    "__version__",
    "Lock",
    "RateLimiter",
    "Semaphore",
    "SharedResourcePool",
    "RedGuard",
]



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager
from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from redis.asyncio import Redis

logger = getLogger(__name__)


class SynchronizationPrimitive(AbstractAsyncContextManager, ABC):
    def __init__(self, client: Redis, name: str) -> None:
        self.client = client
        self.name = name

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    async def _blocking_acquire(self) -> None:
        while True:
            acquired = await self.maybe_acquire()
            if acquired:
                return
            await asyncio.sleep(0.1)

    async def acquire(
        self, *, blocking: bool = False, timeout: float | None = None
    ) -> bool:
        if blocking:
            await asyncio.wait_for(self._blocking_acquire(), timeout=timeout)
            return True
        else:
            return await self.maybe_acquire()

    @abstractmethod
    async def release(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def maybe_acquire(self) -> bool:
        raise NotImplementedError

    async def __aenter__(self) -> bool:
        logger.debug("Acquiring %s", self)
        return await self.acquire(blocking=True)

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        logger.debug("Releasing %s", self)
        await self.release()
